#include "nethost.h"
#include "net64.h"
#include "stringops.h"
#include "argstream.h"
#include <stdio.h>

#ifdef _DEBUG
#ifdef WIN32
#define __DUMPMEMLEAKS__
#endif
#endif

#ifdef __DUMPMEMLEAKS__
#include <crtdbg.h>       // MSVC function: _CrtDumpMemoryLeaks
#endif

#ifdef WIN32
	WSADATA WSAData;
#endif

enum {
	ACMD_NULL = 0,
	ACMD_WRITE,
	ACMD_WRTADR,
	ACMD_WRTBIN,
	ACMD_FILL,
	ACMD_JUMP,
	ACMD_RUN,
	ACMD_EXECUTE
};

enum {
	ATOKEN_NULL = 0,
	ATOKEN_IP,
	ATOKEN_PORT,
	ATOKEN_TIMEOUT,
	ATOKEN_WRITE,
	ATOKEN_WRTADR,
	ATOKEN_WRTBIN,
	ATOKEN_EXECUTE,
	ATOKEN_FILL,
	ATOKEN_JUMP,
	ATOKEN_RUN
};

static ArgToken Tokens[11]=
{
	{ "-n", ATOKEN_IP      },
	{ "-p", ATOKEN_PORT    },
	{ "-t", ATOKEN_TIMEOUT },
	{ "-w", ATOKEN_WRITE   },
	{ "-wa",ATOKEN_WRTADR  },
	{ "-wb",ATOKEN_WRTBIN  },
	{ "-x", ATOKEN_EXECUTE },
	{ "-f", ATOKEN_FILL    },
	{ "-e", ATOKEN_JUMP    },
	{ "-r", ATOKEN_RUN     },
	{ 0,    ATOKEN_NULL    }
};

static void printversion(void)
{
#if defined(__TIME__) && defined(__DATE__)
	printf("Compiled on %s, %s\n", __DATE__, __TIME__);
#endif

#if defined(_MSC_VER)
	printf("MSC %i\n", _MSC_VER);
#endif

#if defined(__INTEL_COMPILER)
	printf("INTEL_COMPILER %i\n", __INTEL_COMPILER);
#endif

#if defined(__WATCOMC__)
	printf("WATCOMC %i\n", __WATCOMC__);
#endif

#if defined(__GNUC__) && defined(__GNUC_MINOR__)
	#if defined(__GNUC_PATCHLEVEL__)
		#if defined(__CYGWIN__)
			printf("CYGWIN GNUC %i.%i.%i\n", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
		#elif defined(__MINGW32__)
			printf("MINGW32 GNUC %i.%i.%i\n", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
		#else
			printf("GNUC %i.%i.%i\n", __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
		#endif
	#else
		#if defined(__CYGWIN__)
			printf("CYGWIN GNUC %i.%i.%i\n", __GNUC__, __GNUC_MINOR__);
		#elif defined(__MINGW32__)
			printf("MINGW32 GNUC %i.%i.%i\n", __GNUC__, __GNUC_MINOR__);
		#else
			printf("GNUC %i.%i.%i\n", __GNUC__, __GNUC_MINOR__);
		#endif
	#endif
#endif

#if defined(__SASC) && defined(__VERSION__) && defined(__REVISION__)
	printf("SASC %i.%i\n", __VERSION__, __REVISION__);
#else
	#if defined(__SASC__)
		printf("SASC %i\n", __SASC__);
	#endif
#endif
}

static int main_command(uBYTE cmd, const char *ip, uWORD port, uLONG timeout,
				 const char *filename, uWORD address, uWORD endaddr, uBYTE fill)
{
	FILE *file;
	NetSocket socket;
	NetHost host;
	Net64 c64;
	int retval, count;
	uLONG size;
	uBYTE buffer[160];

#ifdef _DEBUG
	printf("main_command(%i, %s, %i, %i, %s, 0x%04X, 0x%04X, 0x%02X);\n", cmd, ip, port, timeout, filename, address, endaddr, fill);
#endif

	socket.Init();
	host.Init(&socket, ip, port);
	c64.Init(&host);
	c64.SetTimeout(timeout);

	retval = 0;
	switch (cmd) {

	case ACMD_WRITE:
	case ACMD_WRTADR:
	case ACMD_WRTBIN:
	case ACMD_EXECUTE:
		file = fopen(filename, "rb");

		if (file) {
#ifdef WIN32
			LARGE_INTEGER freq, time1, time2;
			double dfreq, dtime1, dtime2;

			QueryPerformanceFrequency(&freq);
			dfreq = (double)freq.QuadPart;
			QueryPerformanceCounter(&time1);
#endif

			if (cmd != ACMD_WRTBIN) {
				size = fread((void *)buffer, 1, 2, file);

				if (cmd != ACMD_WRTADR) {
					address  = buffer[1];
					address <<= 8;
					address |= buffer[0];
				}
			}

			endaddr = address;

			count = 0;
			while ((!feof(file)) && (retval >= 0)) {

				size = fread((void *)buffer, 1, 128, file);
				if (size > 0) {

					retval = c64.SendData(endaddr, (uWORD)size, buffer);
					if (count == 0) {

						if (retval >= 0) {
							printf("Transferring file %s to $%04X\n", filename, address);
							count++;
						} else {
							printf("C64 did not acknowledge file transfer request!\n");
						}
					}
				}
				endaddr += (uWORD)size;
			}

			if ((count > 0) && (retval < 0)) {
				printf("Transfer interruptet!\n");
			}

			if ((retval >= 0) && (cmd == ACMD_EXECUTE)) {
				if (address > 0x0801) {
					retval = c64.ExecJump(address);
				} else {
					retval = c64.ExecRun();
				}
			}

#ifdef WIN32
			if (retval >= 0) {
				QueryPerformanceCounter(&time2);
				dtime1 = (double)time1.QuadPart;
				dtime2 = (double)time2.QuadPart;
				printf("Transfer speed: %0.2f KiB/s\n", ((float)(((double)(endaddr-address))/((dtime2-dtime1)/dfreq)))/1024.0f);
			}
#endif
		} else {
			printf("File not found.\n");
		}

		break;

	case ACMD_FILL:
		retval = c64.SendFill(address, endaddr-address, fill);
		if (retval >= 0) {
			printf("Filling memory from $%04X to $%04X with $%02X.\n", address, endaddr, fill);
		} else {
			printf("C64 did not acknowledge memory fill request!\n");
		}
		break;

	case ACMD_JUMP:
		retval = c64.ExecJump(address);
		if (retval >= 0) {
			printf("Jumping to address $%04X.\n", address);
		} else {
			printf("C64 did not acknowledge jump request!\n");
		}
		break;

	case ACMD_RUN:
		retval = c64.ExecRun();
		if (retval >= 0) {
			printf("Executing Basic program.\n");
		} else {
			printf("C64 did not acknowledge Basic execute request!\n");
		}
		break;
	};

	c64.Exit();
	host.Exit();
	socket.Exit();

	return retval;
}

int main(int argc, char *argv[])
{
	char *ip;
	int retval;
	char *filename;
	uBYTE cmd, fill;
	uWORD address, endaddr, port;
	unsigned long timeout;
	ArgStream stream;

	stream.Init(argv, argc, Tokens);

	filename = 0;
	address = 0;
	endaddr = 0;
	cmd = ACMD_NULL;
	fill = 0;

	ip = (char*)"192.168.1.64";
	port = 6462;
	timeout = 500;

	stream.GetNext();
	while (stream.HasArguments()) {

		switch (stream.GetNextToken()) {

		case ATOKEN_IP:
			if (stream.HasArguments())
				ip = stream.GetNext();
			break;

		case ATOKEN_PORT:
			if (stream.HasArguments())
				port = (uWORD)ConvertNumber(stream.GetNext());
			break;

		case ATOKEN_TIMEOUT:
			if (stream.HasArguments())
				timeout = ConvertNumber(stream.GetNext());
			break;

		case ATOKEN_WRITE:
			if (stream.HasArguments()) {
				filename = stream.GetNext();
				cmd = ACMD_WRITE;
			}
			break;

		case ATOKEN_WRTADR:
			if (stream.HasArguments()) {
				filename = stream.GetNext();

				if (stream.HasArguments()) {
					address = (uWORD)ConvertNumber(stream.GetNext());
					cmd = ACMD_WRTADR;
				}
			}
			break;

		case ATOKEN_WRTBIN:
			if (stream.HasArguments()) {
				filename = stream.GetNext();

				if (stream.HasArguments()) {
					address = (uWORD)ConvertNumber(stream.GetNext());
					cmd = ACMD_WRTBIN;
				}
			}
			break;

		case ATOKEN_EXECUTE:
			if (stream.HasArguments()) {
				filename = stream.GetNext();
				cmd = ACMD_EXECUTE;
			}
			break;

		case ATOKEN_FILL:
			if (stream.HasArguments()) {
				address = (uWORD)ConvertNumber(stream.GetNext());

				if (stream.HasArguments()) {
					endaddr = (uWORD)ConvertNumber(stream.GetNext());

					if (stream.HasArguments()) {
						fill = (uBYTE)ConvertNumber(stream.GetNext());
						cmd = ACMD_FILL;
					}
				}
			}
			break;

		case ATOKEN_JUMP:
			if (stream.HasArguments()) {
				address = (uWORD)ConvertNumber(stream.GetNext());
				cmd = ACMD_JUMP;
			}
			break;

		case ATOKEN_RUN:
			cmd = ACMD_RUN;
			break;

		}
	}

	stream.Exit();

	if (cmd == ACMD_NULL) {
		printf("\nCodeNet V0.3 by John \"Graham\" Selck on 6.11.2010.\n");
		printversion();
		printf("\nOptions:\n");
		printf("    -n IP-address      Sets the IP-address of the C64 server.\n");
		printf("    -p Port            Sets the port number of the protocol.\n");
		printf("    -w Filename        Sends a PRG file to the C64.\n");
		printf("    -wa Filename Addr  Sends a PRG file to the C64 to a specified address.\n");
		printf("    -wb Filename Addr  Sends a binary file to the C64 to a specified address.\n");
		printf("    -x Filename        Sends a PRG file to the C64 and executes it\n");
		printf("    -f Start End Fill  Fills a block of C64 memory.\n");
		printf("    -e Adress          Jumps to an address in memory.\n");
		printf("    -r                 Executes a program via \"RUN\".\n");
		printf("\nExample:\n");
		printf("> codenet -f 0x1000 0x2080 0xC0 -n 192.168.1.99\n");

#ifdef __DUMPMEMLEAKS__
		_CrtDumpMemoryLeaks();
#endif

		return 0;
	}

#ifdef WIN32
	if (WSAStartup(MAKEWORD(2, 0), &WSAData) != 0) {
		printf("WSAStartup failed!\n");
		return 1;
	}
#endif

	retval = main_command(cmd, ip, port, timeout, filename, address, endaddr, fill);

#ifdef WIN32
	WSACleanup();
#endif

#ifdef __DUMPMEMLEAKS__
	_CrtDumpMemoryLeaks();
#endif

	return (retval < 0);
}
